/***************************************************************************
                 base64.cpp  --  test file for string checks

    $Id: base64.cpp,v 1.24 2004/06/28 18:07:12 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <ulxmlrpcpp.h>

#include <stdio.h>
#include <iostream>

#include <ulxr_value.h>
#include <ulxr_except.h>


int dtable[256];	      /* Encode / decode table */
bool success = true;

void check_malformed(const ulxr::CppString &s)
{
     try
     {
       ulxr::CppString s0 = ulxr::xmlUnEscape(s); // throw
       ULXR_COUT << ULXR_PCHAR("incorrectly converted malformed xml string: ") << s << std::endl;
       success = false;
     }
     catch(...)
     {
        ULXR_COUT << ULXR_PCHAR("correctly caught exception due to malformed xml string: ") << s << std::endl;
     }
}

int main()
{
  try
  {
     ulxr::Char ulxr_char;
     ULXR_COUT << ULXR_PCHAR("sizeof(<string>): " << sizeof(ulxr_char) << std::endl);

     ulxr::CppString s0;
     for (unsigned ii = 0; ii < 256; ++ii)
       s0 += ii;
//     ULXR_COUT << s0 << std::endl;

     ulxr::CppString regression = 
        ULXR_PCHAR("AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vMDEyMzQ1\r\n")
        ULXR_PCHAR("Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWpr\r\n")
        ULXR_PCHAR("bG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6Ch\r\n")
        ULXR_PCHAR("oqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX\r\n")
        ULXR_PCHAR("2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7/P3+/w==\r\n");


     ULXR_COUT << ULXR_PCHAR("Encode 64\n");
     ulxr::CppString s64 = ulxr::encodeBase64(s0);
     ULXR_COUT << s64 << std::endl;

     if (regression != s64)
     {
        ULXR_COUT << ULXR_PCHAR("Encode64 comparison with regression string failed.\n");
        success = false;
     }
     else
       ULXR_COUT << ULXR_PCHAR("Encode64 comparison with regression string OK.\n");

     ULXR_COUT << ULXR_PCHAR("Decode 64\n");
     ulxr::CppString ss = ulxr::decodeBase64(s64);
//     ULXR_COUT << ss << std::endl;

     if (s0 == ss)
       ULXR_COUT << ULXR_PCHAR("base ok\n");
     else
     {
       ULXR_COUT << ULXR_PCHAR("base NOT ok\n");
       success = false;
     }

     s0 = ULXR_PCHAR("-\r=\t#\n+&<>\"\';");
     ulxr::CppString se = ulxr::xmlEscape(s0);
     ULXR_COUT << ULXR_PCHAR("Escaped: ") << se << std::endl;

     if (s0 == ulxr::xmlUnEscape(se))
       ULXR_COUT << ULXR_PCHAR("escape ok\n");
     else
     {
       ULXR_COUT << ULXR_PCHAR("escape NOT ok\n");
       success = false;
     }

     {
       check_malformed(ULXR_PCHAR("abcd&"));
       check_malformed(ULXR_PCHAR("abcd&#x12"));
       check_malformed(ULXR_PCHAR("abcd&#12"));
       check_malformed(ULXR_PCHAR("abcd&#123456789"));
       check_malformed(ULXR_PCHAR("abcd&#x123456789"));
       check_malformed(ULXR_PCHAR("abcd&#123456789;"));
       check_malformed(ULXR_PCHAR("abcd&#x123456789;"));
     }

#ifdef HAVE_ICONV_H

     ulxr::Cpp8BitString s_lat1 = "ܧܤ{}[]/'ܧܤ{}[]/'";
     ulxr::Cpp8BitString s02 = ulxr::encodingToUtf8(s_lat1, "LATIN1");
     ULXR_COUT << ULXR_PCHAR("s02: ") << ULXR_GET_STRING(s02) << std::endl;
     ULXR_COUT << ULXR_PCHAR("s_lat1: ") << ULXR_GET_STRING(s_lat1) << std::endl;
     ULXR_COUT << ULXR_PCHAR("back to orig: ")
               << ULXR_GET_STRING(ulxr::utf8ToEncoding(s02, "LATIN1"))
               << std::endl;
     if (s_lat1 == ulxr::utf8ToEncoding(s02, "LATIN1"))
       ULXR_COUT << ULXR_PCHAR("encoding ok\n");
     else
       ULXR_COUT << ULXR_PCHAR("encoding NOT ok\n");

#endif

     ulxr::Cpp16BitString ws0;
     ws0.reserve(0xd800);
     for (unsigned i = 0x00; i < 0xd800; ++i)
       ws0 += i;

     ulxr::RpcString str (ws0);
     ss = str.getString();
     ULXR_COUT << ss.length() << std::endl;

     ulxr::Cpp16BitString wss = str.getUnicodeString();

     ULXR_COUT << ws0.length() << std::endl;
     ULXR_COUT << (int) ws0[0] << std::endl;

     ULXR_COUT << wss.length() << std::endl;
     ULXR_COUT << (int) wss[0] << std::endl;

     if (ws0 == wss)
       ULXR_COUT << ULXR_PCHAR("unicode 1 ok\n");
     else
     {
       ULXR_COUT << ULXR_PCHAR("unicode 1 NOT ok\n");
       success = false;
     }

#ifdef ULXR_UNICODE

     ws0 = ULXR_PCHAR("khera>&sd< asd<f>e56&&& \"$");
           ULXR_PCHAR("\"\"\'&<&>sd w45234908 0xd9f0�)/()/")
           ULXR_PCHAR("()N u6we4\"E!\"%&((>YXC>YXASESGYDX>YVBHJE%W&TW$");

     ulxr::RpcString str2 (ws0);
     ss = str2.getString();
     wss = str2.getUnicodeString();

     if (ws0 == wss)
       ULXR_COUT << ULXR_PCHAR("unicode 2 ok\n");
     else
     {
       ULXR_COUT << ULXR_PCHAR("unicode 2 NOT ok\n");
       success = false;
     }

#endif

     ULXR_COUT << ULXR_PCHAR("Ready.\n");
  }
  catch(ulxr::Exception &ex)
  {
     ULXR_COUT << ULXR_PCHAR("Error occured: ")
               << ULXR_GET_STRING(ex.why()) << std::endl;
     success = false;
  }

  ULXR_COUT << ULXR_PCHAR("Terminating.\n");
  return success ? 0 : 1;
}
