/***************************************************************************
               rpc_server.cpp  --  test file for an rpc server

    $Id: rpc_server.cpp,v 1.33 2004/08/11 17:26:48 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

// upon requst you may also use intrinsic types like bool, int, double, char*
// which expand to their xmlrpc counterparts. Define before #including ulxr_value.h!!
#define ULXR_USE_INTRINSIC_VALUE_TYPES

// upon requst you may also use the following construct
//      << ("second" << i2)
#define ULXR_STRUCT_MEMBER_FROM_NAME_VALUE

#include <ulxmlrpcpp.h>  // always first header

#include <iostream>
#include <memory>

#include <ulxr_tcpip_connection.h>  // first, don't move: msvc #include bug
#include <ulxr_ssl_connection.h>  
#include <ulxr_http_protocol.h> 
#include <ulxr_except.h>
#include <ulxr_signature.h>
#include <ulxr_dispatcher.h>

class TestClass
{
 public:
   static ulxr::MethodResponse testcall (const ulxr::MethodCall &calldata);
};


ulxr::MethodResponse TestClass::testcall (const ulxr::MethodCall &/*calldata*/)
{
  ulxr::MethodResponse resp(ulxr::Integer(654321));
  resp.setResult(ulxr::Integer(123456));
  return resp;
}


ulxr::MethodResponse stringcall (const ulxr::MethodCall &calldata)
{
  ulxr::RpcString rpcs = calldata.getParam(0);
  ulxr::CppString s = rpcs.getString();
  ulxr::MethodResponse resp;
  resp.setResult(ulxr::RpcString(s));
  return resp;
}


////////////////////////////////////////////////////////////////////////


class TestWorker
{
 public:

   TestWorker () : running(true)
   {}
     

   ulxr::MethodResponse shutdown (const ulxr::MethodCall &/*calldata*/)
   {
     ULXR_COUT << ULXR_PCHAR("got signal to shut down\n");
     ulxr::MethodResponse resp;
     resp.setResult(ulxr::Boolean(running));
     running = false;
     return resp;
   }

   ulxr::MethodResponse testcall1 (const ulxr::MethodCall &calldata)
   {
     ulxr::MethodResponse resp;
     ulxr::Struct st;
     st.addMember(ULXR_PCHAR("param 1"), calldata.getParam(0));
     return resp;
   }

   ulxr::MethodResponse testcall2 (const ulxr::MethodCall &/*calldata*/)
   {
     ulxr::MethodResponse resp;
     resp.setResult(ulxr::Integer(654321));
     return resp;
   }

   bool running;
};

////////////////////////////////////////////////////////////////////////

static ulxr::MethodResponse testcall (const ulxr::MethodCall &calldata)
{
  ulxr::MethodResponse resp;
  ulxr::Integer i = calldata.getParam(0);
  ulxr::Integer i2 = calldata.getParam(1);

  ulxr::Array ar;
  ar << ulxr::Integer(1) << ulxr::Integer(2) << ulxr::RpcString("3");

// upon requst you may also use intrinsic types like bool, int, double, char*
// which expand to their xmlrpc counterparts. Define before #including ulxr_value.h!!
// #define ULXR_USE_INTRINSIC_VALUE_TYPES
//  todo: ar << 11 << 22 << "33";

  ulxr::Struct st;

  st  << ulxr::make_member(ULXR_PCHAR("Hallo"), ar)
      << ulxr::make_member(ULXR_PCHAR("Hallo wstring"), ulxr::RpcString(L"wstring"))
      << ulxr::make_member(ULXR_PCHAR("before"), i)

// upon requst you may also use the following construct
//      << ("second" << i2)
// #define ULXR_STRUCT_MEMBER_FROM_NAME_VALUE

      << ulxr::make_member(ULXR_PCHAR("num_param"), ulxr::Integer(123))
      << ulxr::make_member(ULXR_PCHAR("after"), ulxr::Integer(i.getInteger()+1111))
      << ulxr::make_member(ULXR_PCHAR("difference"), ulxr::RpcString("1111"));

  resp.setResult(st);
  return ulxr::MethodResponse (st);
}


////////////////////////////////////////////////////////////////////////


int main(int argc, char **argv)
{
  ulxr::CppString host = ULXR_PCHAR("localhost");
  if (argc > 1)
    host = ULXR_GET_STRING(argv[1]);

  unsigned port = 32000;
  if (argc > 2)
    port = ulxr_atoi(argv[2]);

  bool secure = false;
  ulxr::CppString sec = ULXR_PCHAR("unsecured");
  if (argc > 3)
    secure = 0 != ulxr_atoi(argv[3]);
  
  if (secure)  
    sec = ULXR_PCHAR("secured");

  ULXR_COUT << ULXR_PCHAR("Serving ") << sec << ULXR_PCHAR(" rpc requests at ") 
            << host << ULXR_PCHAR(":") << port << std::endl;

  std::auto_ptr<ulxr::TcpIpConnection> conn;
#ifdef ULXR_INCLUDE_SSL_STUFF
  if (secure)
  {
    ulxr::SSLConnection *ssl = new ulxr::SSLConnection (true, host, port);
    ssl->setCryptographyData("password", "foo-cert.pem", "foo-cert.pem");
    conn.reset(ssl);
  }
  else
#endif
#ifdef _MSC_VER
  {
   	  std::auto_ptr<ulxr::TcpIpConnection> temp(new ulxr::TcpIpConnection (true, host, port));
	  conn = temp;
  }
#else
    conn.reset(new ulxr::TcpIpConnection (true, host, port));
#endif
	
  ulxr::HttpProtocol prot(conn.get(), conn->getHostName());
  ulxr::Dispatcher server(&prot);

  try
  {
    server.addMethod(ulxr::make_method(TestClass::testcall),
                     ulxr::Struct::getValueName(),
                     ULXR_PCHAR("testcall_in_class_static"),
                     ULXR_PCHAR(""),
                     ULXR_PCHAR("Testcase with a static method in a class"));

    server.addMethod(ulxr::make_method(testcall),
                     ulxr::Signature() << ulxr::Integer(),
                     ULXR_PCHAR("testcall_function"),
                     ulxr::Signature() << ulxr::Integer()
                                       << ulxr::Integer(),
                     ULXR_PCHAR("Testcase with a c-function"));

    server.addMethod(ulxr::make_method(stringcall),
                     ulxr::Signature() << ulxr::RpcString(),
                     ULXR_PCHAR("stringcall"),
                     ulxr::Signature() << ulxr::RpcString(),
                     ULXR_PCHAR("Testcase return input string"));

    TestWorker worker;

    server.addMethod(ulxr::make_method(worker, &TestWorker::testcall1),
                     ulxr::Signature(ulxr::Struct()),
                     ULXR_PCHAR("testcall_in_class_dynamic"),
                     ulxr::Signature(ulxr::Integer()),
                     ULXR_PCHAR("Testcase with a dynamic method in a class"));

    server.addMethod(ulxr::make_method(worker, &TestWorker::shutdown),
                     ulxr::Signature(ulxr::Boolean()),
                     ULXR_PCHAR("testcall_shutdown"),
                     ulxr::Signature(),
                     ULXR_PCHAR("Testcase with a dynamic method in a class, shut down server, return old state"));

    while (worker.running)
    {
      ulxr::MethodCall call = server.waitForCall();
      ulxr::MethodResponse resp = server.dispatchCall(call);
      if (!prot.isTransmitOnly())
        server.sendResponse(resp);

      if (!prot.isPersistent())
        prot.close();
    }
  }
  
  catch(ulxr::Exception& ex)
  {
    ULXR_COUT << ULXR_PCHAR("Error occured: ") << ULXR_GET_STRING(ex.why()) << std::endl;
    if (prot.isOpen())
    {
      try
      {
        ulxr::MethodResponse resp(1, ex.why() );
        if (!prot.isTransmitOnly())
          server.sendResponse(resp);
      }
      catch(...)
      {
        ULXR_COUT << ULXR_PCHAR("error within exception occured\n");
      }
      prot.close();
    }
    return 1;
  }
  ULXR_COUT << ULXR_PCHAR("Well done, Ready.\n");
  return 0;
}
