/***************************************************************************
               ulxr_http_prootocol.cpp  -  http prootocol
                             -------------------
    begin                : Mon May 3 2004
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_http_protocol.cpp,v 1.7 2004/09/02 18:23:46 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

//#define ULXR_DEBUG_OUTPUT


#include <ulxmlrpcpp.h> // always first


#include <cstdio>
#include <ctime>
#include <sys/stat.h>

#include <cstring>

#if defined(__BORLANDC__) || defined (_MSC_VER)
#include <utility>
#endif

#include <ulxr_http_protocol.h>
#include <ulxr_connection.h>
#include <ulxr_except.h>
#include <ulxr_response.h>
#include <ulxr_call.h>


namespace ulxr {


ULXR_EXPORT HttpProtocol::HttpProtocol(Connection *conn, const CppString &hn)
  : Protocol (conn)
  , hostname(hn)
{
  ULXR_TRACE(ULXR_PCHAR("Protocol"));
  init();
}


ULXR_EXPORT HttpProtocol::~HttpProtocol()
{
  ULXR_TRACE(ULXR_PCHAR("~HttpProtocol"));
}


Protocol * ULXR_EXPORT HttpProtocol::detach()
{
  ULXR_TRACE(ULXR_PCHAR("HttpProtocol::detach()"));
  HttpProtocol *clone = new HttpProtocol(*this);
  clone->connection = connection->detach();
  clone->delete_connection = true;
  return clone;  // return previous and running connection
}


void ULXR_EXPORT HttpProtocol::init()
{
  ULXR_TRACE(ULXR_PCHAR("init"));
  headerprops.clear();
  useragent = ULXR_GET_STRING(ULXR_PACKAGE) + ULXR_PCHAR("/") + ULXR_GET_STRING(ULXR_VERSION);
  userTempFields.clear();
}


void ULXR_EXPORT HttpProtocol::resetConnection()
{
  ULXR_TRACE(ULXR_PCHAR("resetConnection"));
  Protocol::resetConnection();
  header_firstline = ULXR_PCHAR("");
  header_buffer = ULXR_PCHAR("");
  headerprops.clear();
//  userTempFields.clear();
}


CppString ULXR_EXPORT HttpProtocol::getHttpProperty(const CppString &in_name) const
{
  ULXR_TRACE(ULXR_PCHAR("getHttpProperty ") << in_name);
  CppString name = in_name;
  makeLower(name);
  header_property::const_iterator it;

  if ((it = headerprops.find(name)) == headerprops.end() )
    throw ConnectionException(NotConformingError,
                              i18n(ULXR_PCHAR("Http property field not available: "))+name, 400);

  return (*it).second;
}


bool ULXR_EXPORT HttpProtocol::hasHttpProperty(const CppString &in_name) const
{
  ULXR_TRACE(ULXR_PCHAR("hasHttpProperty ") << in_name);
  CppString name = in_name;
  makeLower(name);
  return (headerprops.find(name) != headerprops.end() );
}


void ULXR_EXPORT HttpProtocol::parseHeaderLine()
{
  ULXR_TRACE(ULXR_PCHAR("parseHeaderLine"));
  if (header_firstline.length() == 0)
  {
    header_firstline = header_buffer;
    ULXR_DOUT_HTTP(ULXR_PCHAR("firstline: <") << header_firstline << ULXR_PCHAR(">"));
  }
  else
  {
    CppString nm, cont;
    unsigned pos = header_buffer.find(':');
    if (pos == CppString::npos)
    {

      nm = header_buffer;
      cont = ULXR_PCHAR("");
    }
    else
    {
      nm = header_buffer.substr(0, pos);
      cont = header_buffer.substr(pos);
      cont = header_buffer.substr(pos+1);
    }
    makeLower(nm);
    cont = stripWS(cont);
    nm = stripWS(nm);
    headerprops.insert(std::make_pair(nm, cont));
    ULXR_DOUT_HTTP(ULXR_PCHAR("headerprop: <") << nm
                   << ULXR_PCHAR("> + <") << cont << ULXR_PCHAR("> "));
  }
  header_buffer = ULXR_PCHAR("");
}


Protocol::State ULXR_EXPORT
  HttpProtocol::connectionMachine(char * &buffer, long &len)
{
/*
  Each invokation of this state machine tries to parse one single
  line of the buffer. If the content of the buffer is too small
  (no linefeed found) the content is cached in an internal string
  and used the next time.

  buffer points to the beginning of the next line at return if
  a linefeed has been found. In the message body nothing is done.
*/
  ULXR_TRACE(ULXR_PCHAR("connectionMachine"));
  if (len == 0 || buffer == 0)
    return connstate;

  while (len > 0)
  {
    switch (connstate)
    {
      case ConnStart:
        header_buffer = ULXR_PCHAR("");
        header_firstline = ULXR_PCHAR("");
        connstate = ConnHeaderLine;
      break;

      case ConnPendingCR:
        if (*buffer == '\n') // CR+LF
        {
          --len;
          ++buffer;
        }

        if (header_buffer.length() == 0)
          connstate = ConnSwitchToBody;
        else
          connstate = ConnPendingHeaderLine;
      break;

      case ConnPendingHeaderLine:
        if (header_buffer.length() == 0)
          connstate = ConnSwitchToBody;

        else if (*buffer != ' ')  // continuation line of current header field?
        {
          parseHeaderLine();
          connstate = ConnHeaderLine;
        }
        else
          connstate = ConnHeaderLine;
      break;

      case ConnHeaderLine:
        if (*buffer == '\r')
          connstate = ConnPendingCR;

        else if (*buffer == '\n')
        {
          if (header_buffer.length() == 0)
            connstate = ConnSwitchToBody;
          else
            connstate = ConnPendingHeaderLine;
        }

        else
          header_buffer += *buffer;

        ++buffer;
        --len;
      break;

      case ConnSwitchToBody:
        return connstate = ConnBody;
      /* break; */

      case ConnBody:
        return ConnBody;  // body reached: do nothing more here
      /*break; */

      case ConnError:
        return ConnError;
      /*break; */

      default:
        connstate = ConnError;
        throw ConnectionException(SystemError, i18n(ULXR_PCHAR("connectionMachine(): unknown state")), 500);
    }

    if (connstate == ConnSwitchToBody)
    {
      if (hasHttpProperty(ULXR_PCHAR("content-length")))
      {
        determineContentLength();

        ULXR_TRACE(ULXR_PCHAR("content_length: ") << content_length);
        ULXR_TRACE(ULXR_PCHAR("len: ") << len);

        if (content_length >= 0)
          remain_content_length = content_length - len;
      }
      return connstate;
    }

  }

  return connstate;
}


long ULXR_EXPORT HttpProtocol::getBytesToRead() const
{
  ULXR_TRACE(ULXR_PCHAR("getBytesToRead " << remain_content_length));
  return remain_content_length;
}


void ULXR_EXPORT HttpProtocol::determineContentLength()
{
  ULXR_TRACE(ULXR_PCHAR("determineContentLength"));
  header_property::iterator it;
  if ((it = headerprops.find(ULXR_PCHAR("content-length"))) != headerprops.end() )
  {
    ULXR_TRACE(ULXR_PCHAR(" content-length: ") << (*it).second);
    content_length = ulxr_atoi(getLatin1((*it).second).c_str());
    ULXR_TRACE(ULXR_PCHAR(" length: ") << content_length);
  }
  else
    throw ConnectionException(NotConformingError,
                        i18n(ULXR_PCHAR("Content-Length of message required")), 411);

  remain_content_length = content_length;
  ULXR_TRACE(ULXR_PCHAR(" remain_content_length: ") << content_length);
}


void ULXR_EXPORT
  HttpProtocol::sendResponseHeader(int code,
                                   const CppString &phrase,
                                   const CppString &type, unsigned len,
                                   bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("sendResponseHeader"));
  char stat[40];
  ulxr_sprintf(stat, "%d", code );

  char contlen[40];
  ulxr_sprintf(contlen, "%d", len );

  CppString ps = phrase;

  unsigned pos = 0;
  while ((pos = ps.find('\n', pos)) != CppString::npos)
  {
     ps.replace(pos, 1, ULXR_PCHAR(" "));
     pos += 1;
  }

  pos = 0;
  while ((pos = ps.find(ULXR_CHAR('\r'), pos)) != CppString::npos)
  {
     ps.replace(pos, 1, ULXR_PCHAR(" "));
     pos += 1;
  }

  CppString http_str = (CppString) ULXR_PCHAR("HTTP/1.1 ")
                       + ULXR_GET_STRING(stat) + ULXR_PCHAR(" ") + ps
                       + ULXR_PCHAR("\r\n");

  if (!isPersistent())
    http_str += ULXR_PCHAR("Connection: Close\r\n");

  http_str += ULXR_PCHAR("Content-Length: ") + ULXR_GET_STRING(contlen) + ULXR_PCHAR("\r\n");

  if (len != 0 && type.length() != 0)
    http_str  += ULXR_PCHAR("Content-Type: ") + type + ULXR_PCHAR("\r\n");

  for (unsigned i = 0; i < userTempFields.size(); ++i)
    http_str += userTempFields[i] + ULXR_CHAR("\r\n");
  userTempFields.clear();

  if (!wbxml_mode)
  {
    http_str += ULXR_PCHAR("X-Powered-By: ") + getUserAgent() + ULXR_PCHAR("\r\n")
              + ULXR_PCHAR("Server: ") + hostname + ULXR_PCHAR("\r\n")
              + ULXR_PCHAR("Date: ") + getDateStr() + ULXR_PCHAR("\r\n");
  }

  http_str += ULXR_PCHAR("\r\n");    // empty line at end of header

  ULXR_DOUT_HTTP(ULXR_PCHAR("resp: \n") << http_str.c_str());
#ifdef ULXR_UNICODE
  Cpp8BitString utf = unicodeToUtf8(http_str);
  writeRaw(utf.c_str(), utf.length());
#else
  writeRaw(http_str.c_str(), http_str.length());
#endif
}


void ULXR_EXPORT
  HttpProtocol::sendRequestHeader(const CppString &method,
                                    const CppString &resource,
                                    const CppString &type, unsigned len,
                                    bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("sendRequestHeader"));
  char contlen[40];
  ulxr_sprintf(contlen, "%d", len );

  CppString http_str = method + ULXR_PCHAR(" ") + resource + ULXR_PCHAR(" HTTP/1.1\r\n");

  if(!wbxml_mode)
    http_str += ULXR_PCHAR("User-Agent: ") + getUserAgent() + ULXR_PCHAR("\r\n")
              + ULXR_PCHAR("Host: ") + hostname + ULXR_PCHAR("\r\n");

  http_str += ULXR_PCHAR("Content-Length: ") + ULXR_GET_STRING(contlen) + ULXR_PCHAR("\r\n");

  if (!isPersistent())
    http_str += ULXR_PCHAR("Connection: Close\r\n");

  if (len != 0 && type.length() != 0)
    http_str += ULXR_PCHAR("Content-Type: ") + type + ULXR_PCHAR("\r\n");

  for (unsigned i = 0; i < userTempFields.size(); ++i)
    http_str += userTempFields[i] + ULXR_CHAR("\r\n");
  userTempFields.clear();

  if(!wbxml_mode)
    http_str += ULXR_PCHAR("Date: ") + getDateStr() + ULXR_PCHAR("\r\n");

  http_str += ULXR_PCHAR("\r\n");    // empty line at end of header

  ULXR_DOUT_HTTP(ULXR_PCHAR("req: \n") << http_str.c_str());
#ifdef ULXR_UNICODE
  Cpp8BitString utf = unicodeToUtf8(http_str);
  writeRaw(utf.c_str(), utf.length());
#else
  writeRaw(http_str.c_str(), http_str.length());
#endif
};


CppString ULXR_EXPORT HttpProtocol::getDateStr() const
{
  ULXR_TRACE(ULXR_PCHAR("getDateStr"));
  time_t tm;
  time(&tm);
  char * ct = ulxr_ctime(&tm);
  CppString s = ULXR_GET_STRING(ct);  // "\n" already included!
  s.erase(s.length()-1);              // remove it
  return s;
}


void ULXR_EXPORT
  HttpProtocol::sendNegativeResponse(int status,
                                     const CppString &phrase,
                                     const CppString &info)
{
  ULXR_TRACE(ULXR_PCHAR("sendNegativeResponse"));

  char stat[40];
  ulxr_sprintf(stat, "%d", status );

  CppString msg = i18n(ULXR_PCHAR("<html>")
               ULXR_PCHAR("<head><title>Error occured</title></head>")
               ULXR_PCHAR("<body>")
               ULXR_PCHAR("<b>Sorry, error occured: ")) + ULXR_GET_STRING(stat)
              + ULXR_PCHAR(", ") + phrase;

  if (info.length() != 0)
    msg += ULXR_PCHAR("<br />") + info;

  msg += i18n(ULXR_PCHAR("</b>")
              ULXR_PCHAR("<hr /><p>")
              ULXR_PCHAR("This cute little server is powered by")
              ULXR_PCHAR(" <a href=\"http://ulxmlrpcpp.sourceforge.net\">"));

  msg += ULXR_GET_STRING(ULXR_PACKAGE)
         + ULXR_PCHAR("/v") + ULXR_GET_STRING(ULXR_VERSION)
         + ULXR_PCHAR("</a>")
         + ULXR_PCHAR("</p>")
           ULXR_PCHAR("</body>")
           ULXR_PCHAR("</html>");

  sendResponseHeader(status, phrase, ULXR_PCHAR("text/html"), msg.length());
  ULXR_DOUT_RESP(ULXR_PCHAR("msg:\n") << msg);

#ifdef ULXR_UNICODE
  Cpp8BitString utf = unicodeToUtf8(msg);
  writeRaw(utf.c_str(), utf.length());
#else
  writeRaw(msg.c_str(), msg.length());
#endif
}


void ULXR_EXPORT HttpProtocol::sendRpcResponse(const MethodResponse &resp, bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("sendRpcResponse"));
  if (wbxml_mode)
  {
    std::string xml = resp.getWbXml();
    ULXR_DOUT_XML(binaryDebugOutput(xml));
    sendResponseHeader(200, ULXR_PCHAR("OK"), ULXR_PCHAR("application/x-wbxml-ulxr"), xml.length(), wbxml_mode);
    writeRaw(xml.c_str(), xml.length());
  }
  else
  {
    CppString xml = resp.getXml(0)+ULXR_PCHAR("\n");
    sendResponseHeader(200, ULXR_PCHAR("OK"), ULXR_PCHAR("text/xml"), xml.length(), wbxml_mode);
    ULXR_DOUT_XML(xml);
#ifdef ULXR_UNICODE
    Cpp8BitString utf = unicodeToUtf8(xml);
    writeRaw(utf.c_str(), utf.length());
#else
    writeRaw(xml.c_str(), xml.length());
#endif
  }
}


void ULXR_EXPORT HttpProtocol::sendRpcCall(const MethodCall &call,
                                             const CppString &resource,
                                             bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("sendRpcCall"));
  if (wbxml_mode)
  {
    std::string xml = call.getWbXml();
    ULXR_DOUT_XML(binaryDebugOutput(xml));
    sendRequestHeader(ULXR_PCHAR("POST"), resource, ULXR_PCHAR("application/x-wbxml-ulxr"), xml.length(), wbxml_mode);
    writeRaw(xml.c_str(), xml.length());
  }
  else
  {
    CppString xml = call.getXml(0)+ULXR_PCHAR("\n");
    sendRequestHeader(ULXR_PCHAR("POST"), resource, ULXR_PCHAR("text/xml"), xml.length(), wbxml_mode);
    ULXR_DOUT_XML(xml);
#ifdef ULXR_UNICODE
    Cpp8BitString utf = unicodeToUtf8(xml);
    writeRaw(utf.c_str(), utf.length());
#else
    writeRaw(xml.c_str(), xml.length());
#endif
  }
};


bool ULXR_EXPORT HttpProtocol::responseStatus(CppString &phrase) const
{
  ULXR_TRACE(ULXR_PCHAR("responseStatus"));

  CppString s = stripWS(getFirstHeaderLine());
  if (s.length() == 0)
  {
     s = i18n(ULXR_PCHAR("No connection status available"));
     return false;
  }

  unsigned int pos = s.find(' ');
  if (pos != CppString::npos)  // skip version
    s.erase(0, pos+1);
  else
    s = ULXR_PCHAR("");

  CppString stat;
  s = stripWS(s);
  pos = s.find(ULXR_CHAR(' '));
  if (pos != CppString::npos)
  {
    stat = s.substr(0, pos);
    s.erase(0, pos+1);
  }
  else
  {
    stat = s;
    s = ULXR_PCHAR("");
  }

  phrase = stripWS(s);

  return stat == ULXR_PCHAR("200");
}


bool ULXR_EXPORT HttpProtocol::determineClosing(const CppString &http_ver)
{
  ULXR_TRACE(ULXR_PCHAR("determineClosing"));
  if (   http_ver == ULXR_PCHAR("0.9")
      || http_ver == ULXR_PCHAR("1.0"))
  {
    if (hasHttpProperty(ULXR_PCHAR("connection")))
    {
      CppString s = getHttpProperty(ULXR_PCHAR("connection"));
      makeLower(s);
      return !(s == ULXR_PCHAR("keep-alive"));
    }
    return true;  // close by default
  }
  else  // 1.1 and later
  {
    if (hasHttpProperty(ULXR_PCHAR("connection")))
    {
      CppString s = getHttpProperty(ULXR_PCHAR("connection"));
      makeLower(s);
      return (s == ULXR_PCHAR("close"));
    }
    return false;  // keep open by default
  }
  /*return true; avoids warning */
}


bool ULXR_EXPORT HttpProtocol::getUserPass(CppString &user,
                                           CppString &pass) const
{
   ULXR_TRACE(ULXR_PCHAR("getUserPass"));
   user = ULXR_PCHAR("");
   pass = ULXR_PCHAR("");

   if (hasHttpProperty(ULXR_PCHAR("authorization")) )
   {
     CppString auth = getHttpProperty(ULXR_PCHAR("authorization"));

     ULXR_TRACE(ULXR_PCHAR("getUserPass: ") + auth);
     ULXR_TRACE(ULXR_PCHAR("getUserPass: basic?"));

     // only know basic auth
     CppString auth_id = auth.substr(0, 6);
     makeLower(auth_id);
     if (auth_id != ULXR_PCHAR("basic "))
       return false;

     auth.erase(0, 6);
     auth = decodeBase64(auth);
     ULXR_TRACE(ULXR_PCHAR("getUserPass: ':'? ")  + auth);
     unsigned pos = auth.find(':');
     if (pos != CppString::npos)
     {
       user = stripWS(auth.substr(0, pos));
       pass = stripWS(auth.substr(pos+1));
       ULXR_TRACE(ULXR_PCHAR("getUserPass: user=") +user + ULXR_PCHAR(", pass=")+pass);
       return true;
     }
   }

   return false;
}


void ULXR_EXPORT HttpProtocol::rejectAuthentication(const CppString &realm)
{
  ULXR_TRACE(ULXR_PCHAR("rejectAuthentication: ") + realm);
  addOneTimeHttpField(ULXR_PCHAR("WWW-Authenticate"),
                      ULXR_PCHAR("Basic realm=\"") + realm +ULXR_PCHAR("\""));
  sendNegativeResponse(401, ULXR_PCHAR("Authentication required for realm \"")+ realm + ULXR_PCHAR("\""));
}


void ULXR_EXPORT HttpProtocol::addOneTimeHttpField(const CppString &name, const CppString &value)
{
  ULXR_TRACE(ULXR_PCHAR("addOneTimeHttpField: ") + name + ULXR_PCHAR(": ") + value);
  userTempFields.push_back(stripWS(name) + ULXR_PCHAR(": ") + stripWS(value));
}


void ULXR_EXPORT HttpProtocol::setMessageAuthentication(const CppString &user,
                                                          const CppString &pass)
{
  ULXR_TRACE(ULXR_PCHAR("setMessageAuthentication"));
  CppString s = ULXR_PCHAR("Basic ");
  s += encodeBase64(user + ULXR_PCHAR(":") + pass);
  addOneTimeHttpField(ULXR_PCHAR("Authorization"), s);
}


void ULXR_EXPORT HttpProtocol::setTransmitOnly()
{
  ULXR_TRACE(ULXR_PCHAR("setTransmitOnly"));
  addOneTimeHttpField(ULXR_PCHAR("X-TransmitOnly"), ULXR_PCHAR("true"));
}


bool ULXR_EXPORT HttpProtocol::isTransmitOnly()
{
  ULXR_TRACE(ULXR_PCHAR("isTransmitOnly"));
  return hasHttpProperty(ULXR_PCHAR("X-TransmitOnly"))
         && (getHttpProperty(ULXR_PCHAR("X-TransmitOnly")) == ULXR_PCHAR("true"));
}


CppString ULXR_EXPORT HttpProtocol::getProtocolName()
{
  return ULXR_PCHAR("http");
}

};  // namespace ulxr

