/***************************************************************************
                     ulxr_protocol.cpp  -  rpc prootocol
                             -------------------
    begin                : Mon May 3 2004
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_protocol.cpp,v 1.6 2004/06/03 16:57:06 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

//#define ULXR_DEBUG_OUTPUT


#include <ulxmlrpcpp.h>  // always first header

#include <ulxr_connection.h>
#include <ulxr_protocol.h>
#include <ulxr_response.h>
#include <ulxr_call.h>


namespace ulxr {


ULXR_EXPORT Protocol::Protocol(Connection *conn)
  : connection (conn)
  , delete_connection(false)
{
  ULXR_TRACE(ULXR_PCHAR("Protocol"));
  init();
}


ULXR_EXPORT Protocol::~Protocol()
{
  ULXR_TRACE(ULXR_PCHAR("~Protocol"));
  if (delete_connection)
    delete connection;
  connection = 0;
}


void ULXR_EXPORT Protocol::init()
{
  ULXR_TRACE(ULXR_PCHAR("init"));
  setPersistent(false);
  resetConnection();
}


void ULXR_EXPORT Protocol::writeRaw(char const *buff, long len)
{
  ULXR_TRACE(ULXR_PCHAR("writeRaw"));
   connection->write(buff, len);
}


long ULXR_EXPORT Protocol::readRaw(char *buff, long len)
{
  ULXR_TRACE(ULXR_PCHAR("readRaw, want: ") << len);
  if (remain_content_length >= 0)
  {
    ULXR_TRACE(ULXR_PCHAR("read 0 ") << len << ULXR_PCHAR(" ") << remain_content_length);
    if (remain_content_length < len)
      len = remain_content_length;
  }
 
  long readed = connection->read(buff, len);
  
  if (remain_content_length >= 0)
    remain_content_length -= readed;
  
  return readed;
}


void ULXR_EXPORT Protocol::open()
{
  ULXR_TRACE(ULXR_PCHAR("open"));
  connection->open();
  resetConnection();
}


bool ULXR_EXPORT Protocol::isOpen() const
{
  ULXR_TRACE(ULXR_PCHAR("isOpen"));
   return connection->isOpen();
}


void ULXR_EXPORT Protocol::accept()
{
  ULXR_TRACE(ULXR_PCHAR("accept"));
  connection->accept();
  resetConnection();
}


void ULXR_EXPORT Protocol::close()
{
  ULXR_TRACE(ULXR_PCHAR("close"));
  connection->close();
  resetConnection();
}


Connection * ULXR_EXPORT Protocol::getConnection()
{
  ULXR_TRACE(ULXR_PCHAR("getConnection"));
  return connection;
}
  

Protocol::State ULXR_EXPORT
  Protocol::connectionMachine(char * &/*buffer*/, long &/*len*/)
{
  ULXR_TRACE(ULXR_PCHAR("connectionMachine"));
  connstate = ConnBody;
  return ConnBody;
}


void ULXR_EXPORT Protocol::sendRpcResponse(const MethodResponse &resp, bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("sendRpcResponse"));
  if (wbxml_mode)
  {
    std::string xml = resp.getWbXml();
    connection->write(xml.c_str(), xml.length());
  }
  else
  {
    CppString xml = resp.getXml(0)+ULXR_PCHAR("\n");

#ifdef ULXR_UNICODE
    Cpp8BitString utf = unicodeToUtf8(xml);
    connection->write(utf.c_str(), utf.length());
#else
    connection->write(xml.c_str(), xml.length());
#endif
  }
}


void ULXR_EXPORT Protocol::sendRpcCall(const MethodCall &call,
                                       const CppString &/*resource*/,
                                       bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("sendRpcCall"));
  if (wbxml_mode)
  {
    std::string xml = call.getWbXml();
    connection->write(xml.c_str(), xml.length());
  }
  else
  {
    CppString xml = call.getXml(0)+ULXR_PCHAR("\n");
#ifdef ULXR_UNICODE
   Cpp8BitString utf = unicodeToUtf8(xml);
    connection->write(utf.c_str(), utf.length());
#else
    connection->write(xml.c_str(), xml.length());
#endif
  }
};


void ULXR_EXPORT
  Protocol::addAuthentication(const CppString &user,
                              const CppString &pass,
                              const CppString &realm)
{
  ULXR_TRACE(ULXR_PCHAR("addAuthentication"));
  authdata.push_back(AuthData(stripWS(user), stripWS(pass), stripWS(realm)));
}


bool ULXR_EXPORT Protocol::checkAuthentication(const CppString &realm) const
{
  ULXR_TRACE(ULXR_PCHAR("checkAuthentication ") << realm);
  if (authdata.size() == 0)
    return true;   // accept all

  ULXR_TRACE(ULXR_PCHAR("checkAuthentication 1"));
  CppString user, pass;
  if (!getUserPass(user, pass))
    return false;

  ULXR_TRACE(ULXR_PCHAR("checkAuthentication 2 ")
             << ULXR_PCHAR("user: ") << user
             << ULXR_PCHAR(" pass: ") << pass);
  for (unsigned i = 0; i < authdata.size(); ++i)
    if (   authdata[i].user == user
        && authdata[i].pass == pass
        && authdata[i].realm == realm)
      return true;

  ULXR_TRACE(ULXR_PCHAR("checkAuthentication 3"));
  return false;
}


bool ULXR_EXPORT Protocol::getUserPass(CppString & /* user */,
                                       CppString & /* pass */) const
{
  ULXR_TRACE(ULXR_PCHAR("getUserPass"));
  return false;
}


void ULXR_EXPORT Protocol::rejectAuthentication(const CppString & /* realm */)
{
  ULXR_TRACE(ULXR_PCHAR("rejectAuthentication"));
}


void ULXR_EXPORT Protocol::setMessageAuthentication(const CppString & /* user */,
                                                    const CppString & /* pass */)
{
  ULXR_TRACE(ULXR_PCHAR("setMessageAuthentication"));
}


void ULXR_EXPORT Protocol::setTransmitOnly()
{
  ULXR_TRACE(ULXR_PCHAR("setTransmitOnly"));
}


bool ULXR_EXPORT Protocol::isTransmitOnly()
{
  ULXR_TRACE(ULXR_PCHAR("isTransmitOnly"));
  return false; // always return a value
}


void ULXR_EXPORT Protocol::resetConnection()
{
  ULXR_TRACE(ULXR_PCHAR("resetConnection"));
  connstate = ConnStart;
  remain_content_length = -1;
  content_length = -1;
}


};  // namespace ulxr
