/***************************************************************************
               ulxr_tcpip_connection.cpp  -  tcpip connection
                             -------------------
    begin                : Mon Apr 29 2002
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_tcpip_connection.cpp,v 1.33 2004/08/11 17:26:49 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

//#define ULXR_DEBUG_OUTPUT
#define ULXR_REUSE_SOCKET

#ifndef ULXR_OMIT_TCP_STUFF

#include <ulxmlrpcpp.h>  // always first header

#include <cerrno>

#ifdef __SUN__
#include <sys/systeminfo.h>
#endif

#include "ulxr_tcpip_connection.h"
#include "ulxr_except.h"


namespace ulxr {


Connection *ULXR_EXPORT TcpIpConnection::detach()
{
  ULXR_TRACE(ULXR_PCHAR("detach"));
  TcpIpConnection *clone = new TcpIpConnection(*this);
  cut();  
  clone->setServerData(getServerData());
  if (getServerData() != 0)
    getServerData()->incRef();
  return clone;
}


void ULXR_EXPORT TcpIpConnection::init(unsigned prt)
{
  ULXR_TRACE(ULXR_PCHAR("init"));
#if defined(__WIN32__)  && !defined (ULXR_NO_WSA_STARTUP)
  WORD wVersionRequested;
  WSADATA wsaData;
  wVersionRequested = MAKEWORD( 2, 0 );

  if (WSAStartup( wVersionRequested, &wsaData) != 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not initialize Windows sockets: "))
                                + ULXR_GET_STRING(getErrorString(getLastError())), 500);
#endif

  serverdomain = ULXR_PCHAR("");
  remote_name = ULXR_PCHAR("");
  setTimeout(10);
  server_data = 0;
  port = prt;
  hostdata_len = sizeof(hostdata);
  remotedata_len = sizeof(remotedata);
  memset(&hostdata, 0, sizeof(hostdata));
  memset(&remotedata, 0, sizeof(remotedata));
  hostdata.sin_port = htons(port);
  hostdata.sin_family = AF_INET;

  char buffer [1000];
  int ret = gethostname(buffer, sizeof(buffer)-1);
  if (ret != 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not get host name: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  host_name = ULXR_GET_STRING(buffer);

#if defined(__SUN__)

  long status = sysinfo(SI_SRPC_DOMAIN ,buffer, sizeof(buffer)-1);
  if (status == -1)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not get domain name: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  if (buffer[0] != 0)
  {
    host_name += ULXR_PCHAR(".");
    host_name += ULXR_GET_STRING(buffer);
  }

#elif defined(__unix__) || defined(__CYGWIN__)

  ret = getdomainname(buffer, sizeof(buffer)-1);
  if (ret != 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not get domain name: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  if (buffer[0] != 0)
  {
    host_name += ULXR_PCHAR(".");
    host_name += ULXR_GET_STRING(buffer);
  }
  
#elif _WIN32

  struct hostent *hostEntPtr = gethostbyname(getLatin1(host_name).c_str());
  if (!hostEntPtr)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not get host+domain name: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);
  host_name = ULXR_GET_STRING(hostEntPtr->h_name);

#else
# pragma message ("don't know how to determine the domain name")
#endif
}


void ULXR_EXPORT TcpIpConnection::asciiToInAddr(const char *address, struct in_addr &saddr)
{
  memset (&saddr, 0, sizeof(in_addr));
  struct hostent *host;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if ((int)saddr.s_addr == -1)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not perform inet_addr(): "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  host = gethostbyname(address);
  if (host == 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not perform gethostname(): "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  memmove((void*)&saddr, host->h_addr_list, sizeof(in_addr));
}


ULXR_EXPORT TcpIpConnection::TcpIpConnection(bool I_am_server, long adr, unsigned prt)
  : Connection()
{
  ULXR_TRACE(ULXR_PCHAR("TcpIpConnection(bool, long, uint)"));
  init(prt);

  hostdata.sin_addr.s_addr = htonl(adr);

  if (I_am_server)
  {
    server_data = new ServerSocketData(socket(AF_INET, SOCK_STREAM, IPPROTO_TCP));
    if (server_data->getSocket() < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not create socket: "))
                                     + ULXR_GET_STRING(getErrorString(getLastError())), 500);

#ifdef ULXR_REUSE_SOCKET
    int sockOpt = 1;
    if (::setsockopt(server_data->getSocket(), SOL_SOCKET, SO_REUSEADDR,
                     (const char*)&sockOpt, sizeof(sockOpt)) < 0)
      throw ConnectionException(SystemError,
                                      i18n(ULXR_PCHAR("Could not set reuse flag for socket: "))
                                     + ULXR_GET_STRING(getErrorString(getLastError())), 500);
#endif
    
    if((::bind(server_data->getSocket(), (sockaddr*) &hostdata, sizeof(hostdata))) < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not bind adress: "))
                                     + ULXR_GET_STRING(getErrorString(getLastError())), 500);

    listen(server_data->getSocket(), 5);
  }  
}


ULXR_EXPORT
TcpIpConnection::TcpIpConnection(bool I_am_server, const CppString &dom, unsigned prt)
  : Connection()
{
  ULXR_TRACE(ULXR_PCHAR("TcpIpConnection(bool, string, uint)"));
  init(prt);

  unsigned start = 0;
  if (dom.substr(start, 5) == ULXR_PCHAR("http:"))
    start += 5;

  if (dom.substr(start, 2) == ULXR_PCHAR("//"))
    start += 2;

  unsigned slash = dom.find (ULXR_PCHAR("/"), start);
  if (slash != CppString::npos)
    serverdomain = dom.substr(start, slash-1);
  else
    serverdomain = dom;

  struct hostent *hp = gethostbyname(getLatin1(serverdomain).c_str() );
  if (hp == 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Host adress not found: ")) + serverdomain, 500);
  memcpy(&(hostdata.sin_addr), hp->h_addr_list[0], hp->h_length);

  if (I_am_server)
  {
    server_data = new ServerSocketData(socket(AF_INET, SOCK_STREAM, IPPROTO_TCP));
    if (server_data->getSocket() < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not create socket: "))
                                     + ULXR_GET_STRING(getErrorString(getLastError())), 500);
#ifdef ULXR_REUSE_SOCKET
    int sockOpt = 1;
    if (::setsockopt(server_data->getSocket(), SOL_SOCKET, SO_REUSEADDR,
                     (const char*)&sockOpt, sizeof(sockOpt)) < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not set reuse flag for socket: "))
                                     + ULXR_GET_STRING(getErrorString(getLastError())), 500);
#endif
    
    if((::bind(server_data->getSocket(), (sockaddr*) &hostdata, sizeof(hostdata))) < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not bind adress: "))
                                     + ULXR_GET_STRING(getErrorString(getLastError())), 500);

    listen(server_data->getSocket(), 5);
  }  
}


ULXR_EXPORT TcpIpConnection::~TcpIpConnection()
{
  ULXR_TRACE(ULXR_PCHAR("~TcpIpConnection"));
  if (server_data != 0 && server_data->decRef() <= 0)
  {
    ULXR_TRACE(ULXR_PCHAR("delete serverdata"));
    delete server_data;
    server_data = 0;
  }
}


bool ULXR_EXPORT TcpIpConnection::isServerMode() const
{
  return server_data != 0;
}


void ULXR_EXPORT TcpIpConnection::open()
{
  ULXR_TRACE(ULXR_PCHAR("open"));
  if (isOpen() )
    throw RuntimeException(ApplicationError,
                           i18n(ULXR_PCHAR("Attempt to open an already open connection")));

  if (server_data != 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Connection is NOT prepared for client mode")), 500);
//  resetConnection();

  fd_handle = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (fd_handle < 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not create socket: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  if(connect(fd_handle, (struct sockaddr *)&hostdata, sizeof(hostdata)) < 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not connect: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  struct hostent *host;
  remotedata_len = sizeof(remotedata);
  if(getpeername(fd_handle,(struct sockaddr *)&remotedata, &remotedata_len)<0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not get peer data: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);
#ifdef __BORLANDC__
  remote_name = ULXR_PCHAR("<remote-host>");  // FIXME, not working
#else
  else
  {
    if (0 == (host = gethostbyaddr((char*)&remotedata.sin_addr, sizeof remotedata.sin_addr,
                                   AF_INET)))
      remote_name = ULXR_GET_STRING(inet_ntoa(remotedata.sin_addr));
    else                                     
      remote_name = ULXR_GET_STRING(host->h_name);
  }    
#endif
}


void ULXR_EXPORT TcpIpConnection::accept()
{
  ULXR_TRACE(ULXR_PCHAR("accept"));
  if (isOpen() )
    throw RuntimeException(ApplicationError,
                           i18n(ULXR_PCHAR("Attempt to accept an already open connection")));

//  resetConnection();

  if (server_data == 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Connection is NOT prepared for server mode")), 500);


  remotedata_len = sizeof(remotedata);
  if((fd_handle = ::accept(server_data->getSocket(),
                             (sockaddr*) &remotedata, &remotedata_len )) < 0)
    throw ConnectionException(SystemError,
                              i18n(ULXR_PCHAR("Could not accept a connection: "))
                                   + ULXR_GET_STRING(getErrorString(getLastError())), 500);

  struct hostent *host;
  if (0 == (host = gethostbyaddr((char*)&remotedata.sin_addr, sizeof remotedata.sin_addr,
                                 AF_INET)))
    remote_name = ULXR_GET_STRING(inet_ntoa(remotedata.sin_addr));  
  else  
    remote_name = ULXR_GET_STRING(host->h_name);
  ULXR_TRACE(ULXR_PCHAR("/accept"));
}


void ULXR_EXPORT TcpIpConnection::setServerData (ServerSocketData *serv_data)
{
  server_data = serv_data;
}


TcpIpConnection::ServerSocketData * ULXR_EXPORT TcpIpConnection::getServerData () const
{
  return server_data;
}


void ULXR_EXPORT TcpIpConnection::close()
{
  ULXR_TRACE(ULXR_PCHAR("close"));
  remote_name = ULXR_PCHAR("");
  Connection::close();
}


int ULXR_EXPORT TcpIpConnection::abortOnClose(int bOn)
{
  linger sock_linger_struct = {1, 0};
  sock_linger_struct.l_onoff = bOn;

#ifdef __WIN32__
  return setsockopt(server_data->getSocket(), SOL_SOCKET, SO_LINGER,
                    (const char*)&sock_linger_struct, sizeof(linger));
#else                    
  return setsockopt(server_data->getSocket(), SOL_SOCKET, SO_LINGER,
                    &sock_linger_struct, sizeof(linger));
#endif                    
}


CppString ULXR_EXPORT TcpIpConnection::getInterfaceName()
{
  return ULXR_PCHAR("tcpip");
}

int ULXR_EXPORT TcpIpConnection::getLastError()
{
#ifdef __WIN32__
    return(WSAGetLastError());
#else
    return(errno);
#endif
}


};  // namespace ulxr


#endif // ULXR_OMIT_TCP_STUFF
